#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bsp.h"
#include "csg.h"

float csg_vector_dot(csg_vector_t v1, csg_vector_t v2) {
   return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}

void csg_vector_add(csg_vector_t a, csg_vector_t b, csg_vector_t c) {
   a[0] = b[0] + c[0];
   a[1] = b[1] + c[1];
   a[2] = b[2] + c[2];
}

void csg_vector_sub(csg_vector_t a, csg_vector_t b, csg_vector_t c) {
   a[0] = b[0] - c[0];
   a[1] = b[1] - c[1];
   a[2] = b[2] - c[2];
}

void csg_vector_scale(csg_vector_t a, float factor, csg_vector_t b) {
   a[0] = factor * b[0];
   a[1] = factor * b[1];
   a[2] = factor * b[2];
}

void csg_vector_copy(csg_vector_t a, csg_vector_t b) {
   a[0] = b[0];
   a[1] = b[1];
   a[2] = b[2];
}

int csg_plane_side(struct csg_plane_t *p, csg_vector_t v) {
   float    d = csg_vector_dot(p->normal, v) - p->dist;

   if (fabsf(d) < DIST_TOLERANCE)
      return 0;

   if (d < 0.0)
      return -1;

   return 1;
}

struct csg_vertex_t* csg_vertex_new() {
   struct csg_vertex_t        *v;

   v = (struct csg_vertex_t*)malloc(sizeof(struct csg_vertex_t));

   return v;
}

void csg_vertex_delete(struct csg_vertex_t *v) {
   free(v);
}

struct csg_face_t* csg_face_new() {
   struct csg_face_t       *f;

   f = (struct csg_face_t*)malloc(sizeof(struct csg_face_t));
   list_init(&f->vertices);

   return f;
}

void csg_face_delete(struct csg_face_t *f) {
   list_clear(&f->vertices);
   free(f);
}

void csg_face_add_vertices(struct csg_face_t *f, struct csg_vertex_t *v, int n) {
   int            i;

   for (i = 0; i < n; i++)
      list_append(&f->vertices, v + i);
}

struct csg_polytope* csg_polytope_new() {
   struct csg_polytope_t   *p;

   p = (struct csg_polytope_t*)malloc(sizeof(struct csg_polytope_t));

   list_init(&p->vertices);
   list_init(&p->faces);
}

void csg_polytope_delete(struct csg_polytope_t *p) {
   list_rewind(&p->vertices);
   while(list_has_more(&p->vertices))
      free(list_next(&p->vertices));
   list_clear(&p->vertices);
   
   list_rewind(&p->faces);
   while(list_has_more(&p->faces))
      free(list_next(&p->faces));
   list_clear(&p->faces);

   free(p);
}

struct list_item_t* csg_polytope_add_vertex(struct csg_polytope_t *p, struct csg_vertex_t *v) {
   return list_append(&p->vertices, v);
}

void csg_polytope_add_faces(struct csg_polytope_t *p, struct csg_face_t *f, int n) {
   int            i;

   for (i = 0; i < n; i++)
      list_append(&p->faces, f + i);
}

struct csg_vertex_t*  csg_plane_line_intersection(struct csg_plane_t *p, csg_vector_t v1, csg_vector_t v2) {
   struct csg_vertex_t     *v;
   csg_vector_t            dir;
   float                   scale;

   csg_vector_sub(dir, v2, v1);
   scale = (p->dist - csg_vector_dot(p->normal, v1)) / csg_vector_dot(p->normal, dir);

   v = csg_vertex_new();
   csg_vector_scale(v->pos, scale, dir);
   csg_vector_add(v->pos, v->pos, v1);

   return v;
}

int csg_face_split(struct csg_face_t *face, struct csg_plane_t *split_plane, struct csg_face_split_t *split_info) {
   struct list_item_t         *prev_item, *act_item, *ipoint[2];
   struct csg_vertex_t        *prev_vertex, *act_vertex;
   int                        inum;

   // Calculate vertex orientation
   list_rewind(&face->vertices);
   while (list_has_more(&face->vertices)) {
      act_vertex = list_act(&face->vertices);
      act_vertex->side = csg_plane_side(split_plane, act_vertex->pos);

      list_next(&face->vertices);
   }

   // Find edge intersections and insert intersection points.
   inum = 0;
   list_rewind(&face->vertices);
   act_item = list_last_item(&face->vertices);
   act_vertex = list_get_item_data(act_item);
   while (list_has_more(&face->vertices)) {
      prev_item = act_item;
      prev_vertex = act_vertex;

      act_item = list_act_item(&face->vertices);
      act_vertex = list_get_item_data(act_item);

      if (act_vertex->side == 0) {
         // The intersection point is the actual vertex
         ipoint[inum++] = act_item;
         if (inum == 2)
            break;

      } else if (prev_vertex->side * act_vertex->side < 0) {
         // The intersecion point is on the edge
         struct csg_vertex_t     *new_v;

         new_v = csg_plane_line_intersection(split_plane, act_vertex->pos, prev_vertex->pos);

         //csg_polytope_add_vertices(poly, new_v, 1);
         ipoint[inum++] = list_insert_before(&face->vertices, act_item, new_v);
         if (inum == 2)
            break;
      }

      list_next(&face->vertices);
   }

   if (inum < 2) {
      // No intersections found
      split_info->vertex[0] = NULL;
      split_info->vertex[1] = NULL;
      split_info->back_face = NULL;

      return CSG_FACE_SPLIT_ALLNONE;
   }

   list_set_first(&face->vertices, ipoint[0]);
   if (list_last_item(&face->vertices) == ipoint[1])
      list_set_first(&face->vertices, ipoint[1]);

   list_rewind(&face->vertices);
   prev_item = list_act_item(&face->vertices);
   prev_vertex = list_get_item_data(prev_item);
   list_next(&face->vertices);
   act_item = list_act_item(&face->vertices);
   act_vertex = list_get_item_data(act_item);

   switch (act_vertex->side) {
      case -1:
         // Found the back face
         break;

      case 0:
         // The whole edge is on the plane.
         // Look at the rest
         list_next(&face->vertices);
         act_item = list_act_item(&face->vertices);
         act_vertex = list_get_item_data(act_item);
         switch (act_vertex->side) {
            case -1:
               // The whole face belongs to the back side
               split_info->vertex[0] = list_get_item_data(ipoint[0]);
               split_info->vertex[1] = list_get_item_data(ipoint[1]);
               split_info->back_face = NULL;
               
               return CSG_FACE_SPLIT_ALLBACK;

            case 0:
               // The whole face is on the plane
               // No intersections found
               split_info->vertex[0] = NULL;
               split_info->vertex[1] = NULL;
               split_info->back_face = NULL;
               
               return CSG_FACE_SPLIT_ALLPLANE;

           case 1:
               // The whole face belongs to the front side
               split_info->vertex[0] = list_get_item_data(ipoint[0]);
               split_info->vertex[1] = list_get_item_data(ipoint[1]);
               split_info->back_face = NULL;
               
               return CSG_FACE_SPLIT_ALLFRONT;
         }
      
      case 1:
         // Found the front face. Switch intersection points.
         list_set_first(&face->vertices, ipoint[1]);

         list_rewind(&face->vertices);
         prev_item = ipoint[1];
         prev_vertex = list_get_item_data(prev_item);
         list_next(&face->vertices);
         act_item = list_act(&face->vertices);
         act_vertex = list_get_item_data(act_item);
   }

   // Now the list starts with an intersection point
   // and the next vertex is on the back side

   split_info->vertex[0] = list_get_item_data(ipoint[0]);
   split_info->vertex[1] = list_get_item_data(ipoint[1]);
   split_info->back_face = csg_face_new();

   // Add the first intersection point
   list_append(&split_info->back_face->vertices, prev_vertex);
   
   // Add back side vertices to the new face
   // and remove them from the old face
   while (list_has_more(&face->vertices)) {

      act_item = list_act_item(&face->vertices);
      act_vertex = list_get_item_data(act_item);
      
      list_append(&split_info->back_face->vertices, act_vertex);
     
      // Add the second intersection point too
      // but don't remove it from the original face
      if (act_vertex->side == 0)
         break;
      
      list_remove(&face->vertices, act_item);
      list_next(&face->vertices);
   }

   return CSG_FACE_SPLIT_NORMAL;
}

void csg_polytope_split(struct csg_polytope_t **back_poly, struct csg_polytope_t **front_poly, struct csg_plane_t *split_plane) {
   struct list_t              split_edges;

   int                        sides[3] = {0, 0, 0}; // back, plane, front

   // Calculate the orientation of every vertex
   list_rewind(&(*front_poly)->vertices);
   while (list_has_more(&(*front_poly)->vertices)) {
      struct csg_vertex_t        *vertex;
      
      vertex = list_act(&(*front_poly)->vertices);
      vertex->side = csg_plane_side(split_plane, vertex->pos);

      sides[vertex->side + 1]++;

      list_next(&face->vertices);
   }

   if (sides[0] == 0) {
      // No intersection. The whole polytope is on the front or on the plane
      *back_poly = NULL;
      return;

   } else if (sides[2] == 0) {
      // No intersection. The whole polytope is on the back or on the plane
      *back_poly = *front_poly;
      *front_poly = NULL;
      return;
   }

   // There will be a new polytope
   *back_poly = csg_polytope_new();
   // And a list of splitting edges defining a new face
   list_init(&split_edges);

   // Move back side vertices to the new polytope
   list_rewind(&(*front_poly)->vertices);
   while (list_has_more(&(*front_poly)->vertices)) {
      struct list_item_t      *poly_vertex;
      struct csg_vertex_t     *vertex;

      poly_vertex = list_act_item(&(*front_poly)->vertices);
      vertex = list_get_item_data(poly_vertex);

      if (vertex->side == -1) {
         list_unlink(&(*front_poly)->vertices, poly_vertex);
         list_append_item(&(*back_poly)->vertices, poly_vertex);
      }
   }

   // Split faces one-by-one
   list_rewind((*front_poly)->faces);
   for (list_rewind((*front_poly)->faces); list_has_more((*front_poly)->faces); list_next((*front_poly)->faces)) {
      struct list_item_t         *poly_face;
      struct csg_face_t          *front_face, *back_face;
      struct csg_edge_t          *edge;
   
      struct list_item_t         *face_vertex, *face_ipoint[2];
      struct list_item_t         *poly_vertex;
      struct csg_vertex_t        *prev_vertex, *act_vertex;
      int                        inum;

      poly_face = list_act_item((*front_poly)->faces);
      front_face = list_get_item_data(poly_face);

      // Find edge intersections and insert intersection points.
      inum = 0;
      list_rewind(&front_face->vertices);
      face_vertex = list_last_item(&front_face->vertices);
      poly_vertex = list_get_item_data(front_face_vertex);
      act_vertex = list_get_item_data(poly_vertex);

      while (list_has_more(&front_face->vertices)) {
         prev_vertex = act_vertex;

         face_vertex = list_act_item(&front_face->vertices);
         poly_vertex = list_get_item_data(face_vertex);
         act_vertex = list_get_item_data(poly_vertex);

         if (act_vertex->side == 0) {
            // The intersection point is the actual vertex
            face_ipoint[inum++] = face_vertex
            if (inum == 2)
               break;

         } else if (prev_vertex->side * act_vertex->side < 0) {
            // The intersecion point is on the edge
            struct csg_vertex_t     *new_v;

            // Create intersection point
            new_v = csg_plane_line_intersection(split_plane, act_vertex->pos, prev_vertex->pos);

            // Add to the front polytope
            poly_vertex = list_append(&(*front_poly)->vertices, new_v);
            // Add to the front face
            face_ipoint[inum++] = list_insert_before(&front_face->vertices, face_vertex, poly_vertex);
            if (inum == 2)
               break;
         }

         list_next(&front_face->vertices);
      }

      if (inum < 2) {
         // No intersections found. Advance to the next face.
         continue;
      }

      // If the two intersection points are on the same edge then
      // arrange them as the first two points of the list
      list_set_first(&front_face->vertices, face_ipoint[0]);
      if (list_last_item(&front_face->vertices) == face_ipoint[1])
         list_set_first(&front_face->vertices, face_ipoint[1]);

      list_rewind(&front_face->vertices);
      face_vertex = list_act_item(&face->vertices);
      poly_vertex = list_get_item_data(face_vertex);
      prev_vertex = list_get_item_data(poly_vertex);
      
      list_next(&front_face->vertices);
      face_vertex = list_act_item(&face->vertices);
      poly_vertex = list_get_item_data(face_vertex);
      act_vertex = list_get_item_data(poly_vertex);

      // Search for the back face
      switch (act_vertex->side) {
         case -1:
            // Found the back face
            break;

         case 0:
            // The whole edge is on the plane.
            // Look at the rest
            list_next(&front_face->vertices);
            face_vertex = list_act_item(&front_face->vertices);
            poly_vertex = list_get_item_data(face_vertex);
            act_vertex = list_get_item_data(poly_vertex);

            switch (act_vertex->side) {
               case -1:
                  // The whole face belongs to the back side. Only one edge is on the plane.
                  // Add the endpoints of that edge to the back polytope but DON'T remove from the front.
                  // The common vertices will be shared between polytopes.
                  // Remove the face from the front polytope and add to the back.
                 
                  // New split edge
                  edge = csg_edge_new();
                  
                  // Start point
                  edge->front[0] = list_item_get_data(face_ipoint[0]);
                  act_vertex = list_item_get_data(edge->front[0]);
                  edge->back[0] = list_append((*back_poly)->vertices, act_vertex);
                  
                  // End point
                  edge->front[1] = list_item_get_data(face_ipoint[1]);
                  act_vertex = list_item_get_data(edge->front[1]);
                  edge->back[1] = list_append((*back_poly)->vertices, act_vertex);

                  // Add to the split edge list
                  list_append(&split_edges, edge);

                  // Remove the face from the front polytope
                  list_unlink((*front_poly)->faces, poly_face);
                  // Add the face to the back polytope
                  list_append_item((*back_poly)->faces, poly_face);

                  // Advance to the next face
                  continue;

               case 0:
                  // The whole face is on the plane. Theoretically this never happens because in this case
                  // the whole polytope is either on the front or on the back side (because it's convex).
                  continue;

              case 1:
                  // The whole face belongs to the front side
                  
                  // New split edge
                  edge = csg_edge_new();
                  
                  // Start point
                  edge->front[0] = list_item_get_data(face_ipoint[0]);
                  act_vertex = list_item_get_data(edge->front[0]);
                  edge->back[0] = list_append((*back_poly)->vertices, act_vertex);
                  
                  // End point
                  edge->front[1] = list_item_get_data(face_ipoint[1]);
                  act_vertex = list_item_get_data(edge->front[1]);
                  edge->back[1] = list_append((*back_poly)->vertices, act_vertex);

                  // Add to the split edge list
                  list_append(&split_edges, edge);

                  continue;
            }
         
         case 1: {
            // Found the front face. Switch intersection points.
            struct list_item_t         *temp;

            list_set_first(&front_face->vertices, face_ipoint[1]);

            temp = face_ipoint[0];
            face_ipoint[0] = face_ipoint[1];
            face_ipoint[1] = temp;

            list_rewind(&front_face->vertices);
            face_vertex = list_act_item(&face->vertices);
            poly_vertex = list_get_item_data(face_vertex);
            prev_vertex = list_get_item_data(poly_vertex);

            list_next(&front_face->vertices);
            face_vertex = list_act_item(&face->vertices);
            poly_vertex = list_get_item_data(face_vertex);
            act_vertex = list_get_item_data(poly_vertex);
         }
      }

      // Now the list starts with an intersection point
      // and the next vertex is on the back side
 
      // New split edge
      edge = csg_edge_new();
      
      // Start point
      edge->front[0] = list_item_get_data(face_ipoint[0]);
      act_vertex = list_item_get_data(edge->front[0]);
      edge->back[0] = list_append((*back_poly)->vertices, act_vertex);
      
      // End point
      edge->front[1] = list_item_get_data(face_ipoint[1]);
      act_vertex = list_item_get_data(edge->front[1]);
      edge->back[1] = list_append((*back_poly)->vertices, act_vertex);

      // Add to the split edge list
      list_append(&split_edges, edge);
     
      // Create the back face
      back_face = csg_face_new();
      list_append((*back_poly)->faces, back_face);

      // Add the first intersection point to the back face
      list_append(&back_face->vertices, edge->back[0]);

      // Add back side vertices to the back face
      // and remove thrm from the front face
      list_rewind(&front_face->vertices);
      // Skip first intersection point
      list_next(&front_face->vertices);
      while (list_has_more(&face->vertices)) {
         face_vertex = list_act_item(&face->vertices);
         poly_vertex = list_get_item_data(face_vertex);
         act_vertex = list_get_item_data(poly_vertex);

         // Stop at the second intersection point
         if (act_vertex->side == 0)
            break;

         // Remove vertex
         list_unlink(&front_face->vertices, face_vertex);
         // Add to the back face
         list_append(&back_face->vertices, poly_vertex);
        
         list_next(&face->vertices);
      }
      
      // Add the second intersection point to the back face
      list_append(&back_face->vertices, edge->back[1]);
   }
}

