import random

import pygame
from   pygame.locals import *

from   OpenGL.GL import *
from   OpenGL.GLU import *

import config

class Texture:
   def __init__(self, surf):
      w, h = surf.get_size()
      self.surf_size = float(w) / config.VIDEO_MODE[0], float(h) / config.VIDEO_MODE[1]

      w2, h2 = w, h
      if (w2 & (w2-1)) > 0:
         w2 |= w2 >> 1
         w2 |= w2 >> 2
         w2 |= w2 >> 4
         w2 |= w2 >> 8
         w2 |= w2 >> 16
         w2 += 1
         
      if (h2 & (h2-1)) > 0:
         h2 |= h2 >> 1
         h2 |= h2 >> 2
         h2 |= h2 >> 4
         h2 |= h2 >> 8
         h2 |= h2 >> 16
         h2 += 1

      self.tex_size = float(w2) / config.VIDEO_MODE[0], float(h2) / config.VIDEO_MODE[1]
      self.tex_coords = ( float(w) / float(w2), float(h) / float(h2) )

      surf2 = pygame.Surface((w2, h2), 0, surf)
      surf2.blit(surf, (0, 0))

      self.tex_id = glGenTextures(1)
      glBindTexture(GL_TEXTURE_2D, self.tex_id)
      
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
      
      glPixelStorei(GL_UNPACK_ALIGNMENT,1)
      glPixelStorei(GL_PACK_ALIGNMENT, 1)
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w2, h2, 0, GL_RGBA, GL_UNSIGNED_BYTE, pygame.image.tostring(surf2, "RGBA"))

   def __del__(self):
      glDeleteTextures([self.tex_id])

   def getId(self):
      return self.tex_id

   def select(self):
      glBindTexture(GL_TEXTURE_2D, self.tex_id)

   def getSurfaceSize(self):
      return self.surf_size

   def getTextureSize(self):
      return self.tex_size

   def getTextureCoords(self):
      return self.tex_coords

class Image:
   HALIGN_LEFT = 0
   HALIGN_CENTER = 1
   HALIGN_RIGHT = 2

   VALIGN_TOP = 0
   VALIGN_CENTER = 1
   VALIGN_BOTTOM = 2

   def __init__(self, image_file):
      surf = pygame.image.load(image_file).convert_alpha()
      self.texture = Texture(surf)
      self.size = self.texture.getSurfaceSize()

   def getTexture(self):
      return self.texture
   
   def getSize(self):
      return self.size

   def setSize(self, new_size):
      self.size = new_size

   def scale(self, sc):
      self.size = map(lambda x: sc*x, self.size)

   def draw(self, halign = HALIGN_CENTER, valign = VALIGN_CENTER):
      glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT)
      glEnable(GL_TEXTURE_2D)
      glDisable(GL_LIGHTING)
         
      w, h = self.size
      s, t = self.texture.getTextureCoords()

      if halign == Text.HALIGN_LEFT:
         x = 0.0
      elif halign == Text.HALIGN_CENTER:
         x = -w / 2.0
      else:
         x = -w

      if valign == Text.VALIGN_TOP:
         y = 0.0
      elif valign == Text.VALIGN_CENTER:
         y = h / 2.0
      else:
         y = h

      self.texture.select()

      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex2f(x, y)

      glTexCoord2f(0.0, t)
      glVertex2f(x, y - h)

      glTexCoord2f(s, t)
      glVertex2f(x + w, y - h)

      glTexCoord2f(s, 0.0)
      glVertex2f(x + w, y)
      glEnd()

      glPopAttrib()

class Text:
   HALIGN_LEFT = 0
   HALIGN_CENTER = 1
   HALIGN_RIGHT = 2

   VALIGN_TOP = 0
   VALIGN_CENTER = 1
   VALIGN_BOTTOM = 2

   def __init__(self, txt, font_name, font_size):
      font = config.font_manager[font_name][font_size]
      self._createTexture(txt, font)
      self.size = self.texture.getSurfaceSize()

   def _createTexture(self, txt, font):
      surf = font.render(txt, True, (255, 255, 255))
      self.texture = Texture(surf)

   def getTexture(self):
      return self.texture
   
   def getSize(self):
      return self.size

   def setSize(self, new_size):
      self.size = new_size

   def scale(self, sc):
      self.size = map(lambda x: sc*x, self.size)

   def draw(self, halign = HALIGN_CENTER, valign = VALIGN_CENTER):
      glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT)
      glEnable(GL_TEXTURE_2D)
      glDisable(GL_LIGHTING)
         
      w, h = self.size
      s, t = self.texture.getTextureCoords()

      if halign == Text.HALIGN_LEFT:
         x = 0.0
      elif halign == Text.HALIGN_CENTER:
         x = -w / 2.0
      else:
         x = -w

      if valign == Text.VALIGN_TOP:
         y = 0.0
      elif valign == Text.VALIGN_CENTER:
         y = h / 2.0
      else:
         y = h

      self.texture.select()

      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex2f(x, y)

      glTexCoord2f(0.0, t)
      glVertex2f(x, y - h)

      glTexCoord2f(s, t)
      glVertex2f(x + w, y - h)

      glTexCoord2f(s, 0.0)
      glVertex2f(x + w, y)
      glEnd()

      glPopAttrib()

class MonoText:
   HALIGN_LEFT = 0
   HALIGN_CENTER = 1
   HALIGN_RIGHT = 2

   VALIGN_TOP = 0
   VALIGN_CENTER = 1
   VALIGN_BOTTOM = 2

   def __init__(self, font_name, font_size, width = 256):
      self._genTexture(config.font_manager[font_name][font_size], width)

   def _genTexture(self, font, width):
      cell_size = width / 16
      
      surf = pygame.Surface((width, width), SRCALPHA, 32)
      
      a = 0
      for y in range(16):
         for x in range(16):
            s = None
            try:
               s = font.render(chr(a), True, (255, 255, 255))
               ts = s.get_size()
               surf.blit(s, ( x*cell_size + ((cell_size - ts[0]) / 2), y*cell_size + ((cell_size - ts[1]) /2) ))
            except pygame.error:
               pass
            a = a + 1
   
      self.texture = Texture(surf)

   def draw(self, text, size, halign = HALIGN_CENTER, valign = VALIGN_CENTER):
      glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)

      glEnable(GL_TEXTURE_2D)
      glDisable(GL_LIGHTING)
      self.texture.select()

      w, h = size * len(text), size

      if halign == Text.HALIGN_LEFT:
         x = 0.0
      elif halign == Text.HALIGN_CENTER:
         x = -w / 2.0
      else:
         x = -w

      if valign == Text.VALIGN_TOP:
         y = -h
      elif valign == Text.VALIGN_CENTER:
         y = -h / 2.0
      else:
         y = 0.0

      tcell = 1.0 / 16.0
      
      glBegin(GL_QUADS)
      for ch in text:
         chx = (ord(ch) % 16) / 16.0
         chy = (ord(ch) / 16) / 16.0
         
         glTexCoord2f(chx, chy + tcell)
         glVertex2f(x, y)
         
         glTexCoord2f(chx + tcell, chy + tcell)
         glVertex2f(x + size, y)

         glTexCoord2f(chx + tcell, chy)
         glVertex2f(x + size, y + size)
         
         glTexCoord2f(chx, chy)
         glVertex2f(x, y + size)

         x += size

      glEnd()
      
      glPopAttrib()

class MultilineText(Text):
   def __init__(self, txt, font_name, font_size, vgap = 5):
      self.vgap = vgap
      Text.__init__(self, txt, font_name, font_size)

   def _createTexture(self, txt, font):
      w = 0
      h = 0
      surfs = []
      for l in txt.splitlines():
         surf = font.render(l, True, (255, 255, 255))
         surfs.append(surf)

         sw, sh = surf.get_size()
         w = max(w, sw)
         h += sh + self.vgap

      h -= self.vgap
         
      surf = pygame.Surface((w, h), 0, surfs[0])
      y = 0
      for l in surfs:
         surf.blit(l, (0, y))
         y += l.get_height() + self.vgap

      self.texture = Texture(surf)

class SourceBlock:
   def __init__(self, text, angle):
      self.text = text
      self.angle = angle

      w, h = text.getSize()

      x = -w / 2.0
      y = h

      glPushMatrix()
      glLoadIdentity()
      glRotatef(angle, 0.0, 0.0, 1.0)

      if angle > 0.0:
         sy, self.miny, sz = gluProject(x, y - h, 0.0)
         sy, self.maxy, sz = gluProject(x + w, y, 0.0)
      else:
         sy, self.miny, sz = gluProject(x + w, y - h, 0.0)
         sy, self.maxy, sz = gluProject(x, y, 0.0)

      self.miny = float(self.miny) / config.VIDEO_MODE[1] - 0.5
      self.maxy = float(self.maxy) / config.VIDEO_MODE[1] - 0.5
      
      glPopMatrix()

   def getMinY(self):
      return self.miny

   def getMaxY(self):
      return self.maxy

   def draw(self):
      glPushMatrix()
      glRotatef(self.angle, 0.0, 0.0, 1.0)
      self.text.draw(Text.HALIGN_CENTER, Text.VALIGN_BOTTOM)
      glPopMatrix()

class SourceCode:
   def __init__(self, source_file, font_name, font_size, min_lines, max_lines, num_blocks):
      self.font_name = font_name
      self.font_size = font_size
      
      self.lines = config.source_manager[source_file]

      self._genBlocks(num_blocks, min_lines, max_lines, 0)

   def getBlocks(self, num_blocks, min_angle = -10.0, max_angle = 10.0):
      blocks = []
      for i in xrange(num_blocks):
         blocks.append(SourceBlock(random.choice(self.blocks), random.uniform(min_angle, max_angle)))

      return blocks

   def _genBlocks(self, count, min_lines, max_lines, vgap):
      glPushAttrib(GL_TEXTURE_BIT)

      num_lines = len(self.lines)
      self.blocks = []
      for i in xrange(count):
         lines = random.randint(min_lines, max_lines)
         start = random.randint(0, num_lines - lines)
         self.blocks.append(MultilineText("\n".join(self.lines[start:start+lines]), self.font_name, self.font_size, vgap))

      glPopAttrib()

